@@ -5,15 +5,19 @@ from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
 
-import keras as k
-import keras.backend as K
+import tensorflow.python.keras as k
 
-from keras.layers import Dense
-from keras.layers import Concatenate
-from keras.layers import InputSpec
-from keras.constraints import MinMaxNorm
+import tensorflow.python.keras.backend as K
+graph_unique_name = K.get_graph().unique_name
+
+from tensorflow.python.keras.layers import Dense
+from tensorflow.python.keras.layers import Concatenate
+from tensorflow.python.keras.layers import InputSpec
+from tensorflow.python.keras.constraints import MinMaxNorm
 
 from ..utils import to_list
+from ..utils import default_constant_initializer
+from ..utils import default_regularizer
 
 from .functional import Functional
 from .variable import Variable
@@ -61,8 +65,7 @@ class Parameter(Functional):
         )
 
         if len(input_tensors) > 1:
-            lay = Concatenate()
-            lay.name = "conct_" + lay.name.split("_")[-1]
+            lay = Concatenate(name=graph_unique_name('conct'))
             lay_input = lay(input_tensors)
         else:
             lay_input = input_tensors[0]
@@ -74,6 +77,10 @@ class Parameter(Functional):
             layers=to_list(layers)
         )
 
+    @classmethod
+    def get_class(cls):
+        return Functional
+
 
 class ParameterBase(Dense):
     """ Base Parameter class to be used for parameter inversion.
@@ -101,7 +108,7 @@ class ParameterBase(Dense):
             units=1,
             use_bias=True,
             kernel_initializer='zeros',
-            bias_initializer=k.initializers.constant(val),
+            bias_initializer=default_constant_initializer(val),
             kernel_regularizer=None,
             bias_regularizer=None,
             activity_regularizer=None,
@@ -131,7 +138,7 @@ class ParameterBase(Dense):
         self.built = True
 
 
-from keras.constraints import Constraint
+from tensorflow.python.keras.constraints import Constraint
 class MinMax(Constraint):
     """MinMax weight constraint.
 
