@@ -7,12 +7,12 @@ from __future__ import print_function
 
 import os
 
-import keras.backend as K
-import keras as k
+import tensorflow.python.keras.backend as K
+import tensorflow.python.keras as k
 import numpy as np
 
-from keras.models import Model
-from keras.utils import plot_model
+from tensorflow.python.keras.models import Model
+from tensorflow.python.keras.utils.vis_utils import plot_model
 
 from ..utils import unpack_singleton, to_list
 from ..utils import is_variable, is_constraint, is_functional
@@ -126,7 +126,7 @@ class SciModel(object):
         # compile the model.
         model.compile(
             loss=loss_func,
-            optimizer=optimizer,
+            optimizer=optimizer
         )
         # model.train_function = True
 
@@ -198,10 +198,15 @@ class SciModel(object):
               learning_rate=0.001,
               shuffle=True,
               callbacks=None,
+              stop_lr_value=1e-8,
+              reduce_lr_after=None,
+              reduce_lr_min_delta=0.001,
               stop_after=None,
+              stop_loss_value=1e-8,
               save_weights_to=None,
               save_weights_freq=0,
               default_zero_weight=0.0,
+              validation_data=None,
               **kwargs):
         """Performs the training on the model.
 
@@ -232,8 +237,16 @@ class SciModel(object):
             shuffle: Boolean (whether to shuffle the training data).
                 Default value is True.
             callbacks: List of `keras.callbacks.Callback` instances.
-            stop_after: To stop after certain missed epochs.
-                Defaulted to epochs.
+            reduce_lr_after: patience to reduce learning rate or stop after certain missed epochs.
+                Defaulted to epochs max(10, epochs/10).
+            stop_lr_value: stop the training if learning rate goes lower than this value.
+                Defaulted to 1e-8.
+            reduce_lr_min_delta: min absolute change in total loss value that is considered a successful change.
+                Defaulted to 0.001. 
+                This values affects number of failed attempts to trigger reduce learning rate based on reduce_lr_after. 
+            stop_after: To stop after certain missed epochs. Defaulted to total number of epochs.
+            stop_loss_value: The minimum value of the total loss that stops the training automatically. 
+                Defaulted to 1e-8. 
             save_weights_to: (file_path) If you want to save the state of the model (at the end of the training).
             save_weights_freq: (Integer) Save weights every N epcohs.
                 Defaulted to 0.
@@ -249,16 +262,21 @@ class SciModel(object):
             # decay_epochs = epochs if decay_epochs is None else decay_epochs
             # a0 = decay_epochs / (f1-f0)
             # n0 = -a0*f0
+            if reduce_lr_after is None:
+                reduce_lr_after = max([10, epochs/10])
+            if stop_after is None:
+                stop_after = epochs
             callbacks = []
             if isinstance(learning_rate, (type(None), float, int)):
                 lr_rates = 0.001 if learning_rate is None else learning_rate
                 K.set_value(self.model.optimizer.lr, lr_rates)
                 callbacks.append(
-                    k.callbacks.callbacks.ReduceLROnPlateau(
+                    k.callbacks.ReduceLROnPlateau(
                         monitor='loss', factor=0.5,
-                        patience=epochs/10, cooldown=epochs/10,
-                        verbose=1, mode='auto', min_delta=0.001,
-                        min_lr=1e-3*lr_rates
+                        patience=reduce_lr_after, #cooldown=epochs/10,
+                        verbose=1, mode='auto', 
+                        min_delta=reduce_lr_min_delta,
+                        min_lr=0.
                     )
                 )
             elif isinstance(learning_rate, (tuple, list)):
@@ -273,11 +291,12 @@ class SciModel(object):
                     "learning rate: expecting a `float` or a tuple/list of two arrays"
                     " with `epochs` and `learning rates`"
                 )
-            if stop_after is None:
-                stop_after = epochs
             callbacks += [
-                k.callbacks.EarlyStopping(monitor="loss", mode='auto', verbose=1, patience=stop_after),
+                k.callbacks.EarlyStopping(monitor="loss", mode='auto', verbose=1,
+                                          patience=stop_after, min_delta=1e-9),
                 k.callbacks.TerminateOnNaN(),
+                EarlyStoppingByLossVal(stop_loss_value),
+                EarlyStoppingByLearningRate(stop_lr_value)
             ]
         # prepare X,Y data.
         x_true = to_list(x_true)
@@ -345,7 +364,7 @@ class SciModel(object):
             try:
                 self._model.save_weights("{}-start.hdf5".format(save_weights_to))
                 model_file_path = save_weights_to + "-{epoch:05d}-{loss:.3e}.hdf5"
-                model_check_point = k.callbacks.callbacks.ModelCheckpoint(
+                model_check_point = k.callbacks.ModelCheckpoint(
                     model_file_path, monitor='loss', save_weights_only=True, mode='auto',
                     period=10 if save_weights_freq==0 else save_weights_freq,
                     save_best_only=True if save_weights_freq==0 else False
@@ -362,6 +381,7 @@ class SciModel(object):
             batch_size=batch_size,
             shuffle=shuffle,
             callbacks=callbacks,
+            validation_data=validation_data,
             **kwargs
         )
         if save_weights_to is not None:
@@ -474,15 +494,20 @@ class SciModel(object):
             ValueError if anything other than "mse" or "mae" is passed.
         """
         if method in ("mse", "mean_squared_error"):
-            return lambda y_true, y_pred: K.mean(K.square(y_true - y_pred))
+            return lambda y_true, y_pred: K.mean(K.square(y_true - y_pred), axis=-1)
         elif method in ("mae", "mean_absolute_error"):
-            return lambda y_true, y_pred: K.mean(K.abs(y_true - y_pred))
+            return lambda y_true, y_pred: K.mean(K.abs(y_true - y_pred), axis=-1)
         elif method in ("se", "squared_error"):
-            return lambda y_true, y_pred: K.sum(K.square(y_true - y_pred))
+            return lambda y_true, y_pred: K.sum(K.square(y_true - y_pred), axis=-1)
         elif method in ("ae", "absolute_error"):
-            return lambda y_true, y_pred: K.sum(K.abs(y_true - y_pred))
+            return lambda y_true, y_pred: K.sum(K.abs(y_true - y_pred), axis=-1)
+        elif hasattr(k.losses, method):
+            return getattr(k.losses, method)
         else:
-            raise ValueError
+            raise ValueError(
+                'Supported losses: Keras loss function or (mse, mae, se, ae)'
+            )
+
 
     @staticmethod
     def _prepare_data(cond_outputs, y_true, global_weights, num_sample, default_zero_weight):
@@ -493,8 +518,19 @@ class SciModel(object):
             ids = None
             yc = cond_outputs[i]
             if isinstance(yt, tuple) and len(yt) == 2:
-                ids = yt[0]
-                ys.append(yt[1])
+                ids = yt[0].flatten()
+                if isinstance(yt[1], np.ndarray):
+                    adjusted_yt = yt[1]
+                    if ids.size == yt[1].shape[0] and ids.size < num_sample:
+                        adjusted_yt = np.zeros((num_sample,)+yt[1].shape[1:])
+                        adjusted_yt[ids, :] = yt[1]
+                    elif yt[1].shape[0] != num_sample:
+                        raise ValueError(
+                            'Error in size of the target {}.'.format(i)
+                        )
+                else:
+                    adjusted_yt = yt[1]
+                ys.append(adjusted_yt)
             elif isinstance(yt, (np.ndarray, str, float, int, type(None))):
                 ys.append(yt)
             else:
@@ -521,9 +557,9 @@ class SciModel(object):
                             'Dimension of expected `y_true` does not match with defined `Constraint`'
                         )
             elif isinstance(ys[-1], str):
-                if ys[-1] == 'zeros':
+                if ys[-1] == 'zero' or ys[-1] == 'zeros':
                     ys[-1] = np.zeros((num_sample, ) + k.backend.int_shape(yc)[1:])
-                elif ys[-1] == 'ones':
+                elif ys[-1] == 'one' or ys[-1] == 'ones':
                     ys[-1] = np.ones((num_sample, ) + k.backend.int_shape(yc)[1:])
                 else:
                     raise ValueError(
@@ -537,5 +573,41 @@ class SciModel(object):
                 raise ValueError(
                     'Unsupported entry - {} '.format(ys[-1])
                 )
+            # set undefined ids to zeros.
+            if ids.size != num_sample:
+                adjusted_ids = np.ones(num_sample, dtype=bool)
+                adjusted_ids[ids] = False
+                ys[-1][adjusted_ids, :] = 0.0
 
         return ys, weis
+
+
+class EarlyStoppingByLossVal(k.callbacks.Callback):
+    def __init__(self, value, stop_after=1):
+        super(k.callbacks.Callback, self).__init__()
+        self.value = value
+        self.wait = stop_after
+        
+    def on_epoch_end(self, epoch, logs={}):
+        current = logs.get('loss')
+        if current < self.value:
+            self.wait -= 1
+            if self.wait <= 0:
+                self.model.stop_training = True
+                print("Epoch {:05d}: early stopping at loss value {:0.6e}".format(epoch, current))
+                print("Revise 'stop_loss_value={:0.12f}' in '.train' if it was not your intent. ".format(self.value))
+
+
+class EarlyStoppingByLearningRate(k.callbacks.Callback):
+    def __init__(self, value):
+        super(k.callbacks.Callback, self).__init__()
+        self.value = value
+
+    def on_epoch_end(self, epoch, logs={}):
+        current = logs.get('lr')
+        if current < self.value:
+            self.model.stop_training = True
+            print("Epoch {:05d}: early stopping at learning rate {:0.6e}".format(epoch, current))
+            print("Revise 'stop_lr_value={:0.12f}' in '.train' if it was not your intent. ".format(self.value))
+
+
