@@ -7,6 +7,8 @@ from __future__ import print_function
 from . import math
 from . import utilities
 from . import validations
+from . import activations
+from . import initializers
 
 from .utilities import *
 from .validations import *
