@@ -5,41 +5,53 @@ from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
 
+import tensorflow.python.keras.backend as K
+graph_unique_name = K.get_graph().unique_name
 
-from keras.layers import Lambda
-from keras.layers import Dot
-from keras.layers import Input
-from keras.models import Model
+from tensorflow.python.keras.layers import Lambda
+from tensorflow.python.keras.layers import Dot
+from tensorflow.python.keras.layers import Input
+from tensorflow.python.keras.models import Model
+
+from tensorflow import gradients as tf_gradients
+from tensorflow import stop_gradient as tf_stop_gradient
 
 from .utilities import *
 from .validations import *
 
 
-def pow(f, a):
+def pow(a, b):
     """Element-wise exponentiation applied to the `Functional` object.
 
     # Arguments
-        f: Functional object.
-        a: Python integer.
+        a, b: pow(a,b)
+        Note that at least one of them should be of type Functional.
 
     # Returns
         A Functional.
     """
-    validate_functional(f)
+    if is_functional(a):
+        validate_functional(a)
+        f, p = a, b
+        name = "pow{:d}".format(p) if isinstance(p, int) else "pow{:.3f}".format(p)
+        lmbd = [Lambda(lambda x: x ** p, name=graph_unique_name(name)) for X in f.outputs]
+
+    elif is_functional(b):
+        validate_functional(b)
+        f, p = b, a
+        name = "pow{:d}".format(p) if isinstance(p, int) else "pow{:.3f}".format(p)
+        lmbd = [Lambda(lambda x: p ** x, name=graph_unique_name(name)) for X in f.outputs]
+
+    else:
+        raise ValueError('Expected one functional in the arguments.')
+
+    Functional = f.get_class()
+    res = Functional(
+        inputs = unique_tensors(f.inputs.copy()),
+        outputs = _apply_operation(lmbd, f),
+        layers = lmbd,
+    )
 
-    res = f.copy()
-    lmbd = [Lambda(lambda x: x**a) for X in f.outputs]
-    for l in lmbd:
-        if isinstance(a, int):
-            l.name = "pow{:d}/".format(a) + l.name.split("_")[-1]
-        elif isinstance(a, float):
-            l.name = "pow{:.3f}/".format(a) + l.name.split("_")[-1]
-        else:
-            raise ValueError(
-                'Expected an `int` or `float` for a in x^a. '
-            )
-    # res.append_to_layers(lmbd)
-    res.outputs = _apply_operation(lmbd, res)
     return res
 
 
@@ -55,18 +67,20 @@ def add(f, other):
     """
     validate_functional(f)
 
-    res = f.copy()
+    inputs = f.inputs.copy()
     if is_functional(other):
-        res.append_to_inputs(other.inputs)
-        res.append_to_layers(other.layers)
-        lmbd = [Lambda(lambda x: x[0]+x[1]) for X in f.outputs]
+        inputs += to_list(other.inputs)
+        lmbd = [Lambda(lambda x: x[0]+x[1], name=graph_unique_name("add")) for X in f.outputs]
     else:
-        lmbd = [Lambda(lambda x: x+other) for X in f.outputs]
-
-    for l in lmbd:
-        l.name = "add/" + l.name.split("_")[-1]
-    # res.append_to_layers(lmbd)
-    res.outputs = _apply_operation(lmbd, res, other)
+        _warn_for_ndarray(other)
+        lmbd = [Lambda(lambda x: x+other, name=graph_unique_name("add")) for X in f.outputs]
+    
+    Functional = f.get_class()
+    res = Functional(
+        inputs = unique_tensors(inputs),
+        outputs = _apply_operation(lmbd, f, other),
+        layers = lmbd
+    )
     return res
 
 
@@ -95,18 +109,20 @@ def sub(f, other):
     """
     validate_functional(f)
 
-    res = f.copy()
+    inputs = f.inputs.copy()
     if is_functional(other):
-        res.append_to_inputs(other.inputs)
-        res.append_to_layers(other.layers)
-        lmbd = [Lambda(lambda x: x[0]-x[1]) for X in f.outputs]
+        inputs += to_list(other.inputs)
+        lmbd = [Lambda(lambda x: x[0]-x[1], name=graph_unique_name("sub")) for X in f.outputs]
     else:
-        lmbd = [Lambda(lambda x: x-other) for X in f.outputs]
+        _warn_for_ndarray(other)
+        lmbd = [Lambda(lambda x: x-other, name=graph_unique_name("sub")) for X in f.outputs]
 
-    for l in lmbd:
-        l.name = "sub/" + l.name.split("_")[-1]
-    # res.append_to_layers(lmbd)
-    res.outputs = _apply_operation(lmbd, res, other)
+    Functional = f.get_class()
+    res = Functional(
+        inputs = unique_tensors(inputs),
+        outputs = _apply_operation(lmbd, f, other),
+        layers = lmbd
+    )
     return res
 
 
@@ -122,18 +138,20 @@ def rsub(f, other):
     """
     validate_functional(f)
 
-    res = f.copy()
+    inputs = f.inputs.copy()
     if is_functional(other):
-        res.append_to_inputs(other.inputs)
-        res.append_to_layers(other.layers)
-        lmbd = [Lambda(lambda x: x[1]-x[0]) for X in f.outputs]
+        inputs += to_list(other.inputs)
+        lmbd = [Lambda(lambda x: x[1]-x[0], name=graph_unique_name("rsub")) for X in f.outputs]
     else:
-        lmbd = [Lambda(lambda x: other-x) for X in f.outputs]
+        _warn_for_ndarray(other)
+        lmbd = [Lambda(lambda x: other-x, name=graph_unique_name("rsub")) for X in f.outputs]
 
-    for l in lmbd:
-        l.name = "rsub/" + l.name.split("_")[-1]
-    # res.append_to_layers(lmbd)
-    res.outputs = _apply_operation(lmbd, res, other)
+    Functional = f.get_class()
+    res = Functional(
+        inputs = unique_tensors(inputs),
+        outputs = _apply_operation(lmbd, f, other),
+        layers = lmbd
+    )
     return res
 
 
@@ -149,19 +167,20 @@ def mul(f, other):
     """
     validate_functional(f)
 
-    res = f.copy()
+    inputs = f.inputs.copy()
     if is_functional(other):
-        res.append_to_inputs(other.inputs)
-        res.append_to_layers(other.layers)
-        lmbd = [Lambda(lambda x: x[0]*x[1]) for X in f.outputs]
+        inputs += to_list(other.inputs)
+        lmbd = [Lambda(lambda x: x[0]*x[1], name=graph_unique_name("mul")) for X in f.outputs]
     else:
-        lmbd = [Lambda(lambda x: x*other) for X in f.outputs]
+        _warn_for_ndarray(other)
+        lmbd = [Lambda(lambda x: x*other, name=graph_unique_name("mul")) for X in f.outputs]
 
-    for l in lmbd:
-        l.name = "mul/" + l.name.split("_")[-1]
-
-    # res.append_to_layers(lmbd)
-    res.outputs = _apply_operation(lmbd, res, other)
+    Functional = f.get_class()
+    res = Functional(
+        inputs = unique_tensors(inputs),
+        outputs = _apply_operation(lmbd, f, other),
+        layers = lmbd
+    )
     return res
 
 
@@ -190,18 +209,20 @@ def div(f, other):
     """
     validate_functional(f)
 
-    res = f.copy()
+    inputs = f.inputs.copy()
     if is_functional(other):
-        res.append_to_inputs(other.inputs)
-        res.append_to_layers(other.layers)
-        lmbd = [Lambda(lambda x: x[0]/x[1]) for X in f.outputs]
+        inputs += to_list(other.inputs)
+        lmbd = [Lambda(lambda x: x[0]/x[1], name=graph_unique_name("div")) for X in f.outputs]
     else:
-        lmbd = [Lambda(lambda x: x/other) for X in f.outputs]
+        _warn_for_ndarray(other)
+        lmbd = [Lambda(lambda x: x/other, name=graph_unique_name("div")) for X in f.outputs]
 
-    for l in lmbd:
-        l.name = "div/" + l.name.split("_")[-1]
-    # res.append_to_layers(lmbd)
-    res.outputs = _apply_operation(lmbd, res, other)
+    Functional = f.get_class()
+    res = Functional(
+        inputs = unique_tensors(inputs),
+        outputs = _apply_operation(lmbd, f, other),
+        layers = lmbd
+    )
     return res
 
 
@@ -217,18 +238,20 @@ def rdiv(f, other):
     """
     validate_functional(f)
 
-    res = f.copy()
+    inputs = f.inputs.copy()
     if is_functional(other):
-        res.append_to_inputs(other.inputs)
-        res.append_to_layers(other.layers)
-        lmbd = [Lambda(lambda x: x[1]/x[0]) for X in f.outputs]
+        inputs += to_list(other.inputs)
+        lmbd = [Lambda(lambda x: x[1]/x[0], name=graph_unique_name("rdiv")) for X in f.outputs]
     else:
-        lmbd = [Lambda(lambda x: other/x) for X in f.outputs]
+        _warn_for_ndarray(other)
+        lmbd = [Lambda(lambda x: other/x, name=graph_unique_name("rdiv")) for X in f.outputs]
 
-    for l in lmbd:
-        l.name = "rdiv/" + l.name.split("_")[-1]
-    # res.append_to_layers(lmbd)
-    res.outputs = _apply_operation(lmbd, res, other)
+    Functional = f.get_class()
+    res = Functional(
+        inputs = unique_tensors(inputs),
+        outputs = _apply_operation(lmbd, f, other),
+        layers = lmbd
+    )
     return res
 
 
@@ -243,20 +266,94 @@ def dot(f, other):
         A Functional.
     """
     validate_functional(f)
+    validate_functional(other)
+    assert len(f.outputs) == len(other.outputs)
 
-    res = f.copy()
-    if is_functional(other):
-        res.append_to_inputs(other.inputs)
-        res.append_to_layers(other.layers)
-        lmbd = [k.layers.Dot(axes=-1) for X in f.outputs]
-    else:
-        lmbd = [Lambda(lambda x: x*other) for X in f.outputs]
+    outputs = []
+    layers = []
+    for fl, fr in zip(f.outputs, other.outputs):
+        assert fl.shape.as_list() == fr.shape.as_list(),\
+            'Expected equal dimensions for output of functionals. '
+        l = Lambda(
+            lambda x: K.reshape(tf.math.reduce_sum(x*fr, list(range(1, len(fl.shape)))), [-1, 1]),
+            name=graph_unique_name("dot")
+        )
+        layers += [l]
+        outputs += [l(fl)]
+        
+    inputs = to_list(f.inputs) + to_list(other.inputs)
+    Functional = f.get_class()
+    res = Functional(
+        inputs = unique_tensors(inputs),
+        outputs = outputs,
+        layers = layers
+    )
+    return res
 
-    for l in lmbd:
-        l.name = "dot/" + l.name.split("_")[-1]
 
-    # res.append_to_layers(lmbd)
-    res.outputs = _apply_operation(lmbd, res, other)
+def diag_part(f):
+    """Diag_part operation returns diagonal part of outputs of (None,N,N) functional.
+
+    # Arguments
+        f: Functional object.
+
+    # Returns
+        A Functional.
+    """
+    validate_functional(f)
+
+    lmbd = []
+    outputs = []
+    for o in f.outputs:
+        assert len(o.shape) == 3, \
+            'Exptected output dimension to be (None, N, N)'
+        dim = o.shape[-1]
+        l = Lambda(
+            lambda x: tf.linalg.diag_part(x),
+            name=graph_unique_name("diag_part")
+        )
+        lmbd += [l]
+        outputs += [l(o)]
+
+    Functional = f.get_class()
+    res = Functional(
+        inputs = f.inputs.copy(),
+        outputs = outputs,
+        layers = lmbd
+    )
+    return res
+
+
+def diag(f):
+    """Diag operation converts a vector output (None, N) to a matrix form of (None,N,N) functional.
+
+    # Arguments
+        f: Functional object.
+
+    # Returns
+        A Functional.
+    """
+    validate_functional(f)
+
+    lmbd = []
+    outputs = []
+    for o in f.outputs:
+        assert len(o.shape) == 2, \
+            'Exptected output dimension to be (None, N)'
+        dim = o.shape[-1]
+        l = Lambda(
+            lambda x: tf.linalg.diag(x),
+            name=graph_unique_name("diag")
+        )
+        lmbd += [l]
+        outputs += [l(o)]
+
+    Functional = f.get_class()
+    res = Functional(
+        inputs = f.inputs.copy(),
+        outputs = outputs,
+        layers = lmbd
+    )
     return res
 
 
@@ -431,6 +528,18 @@ def coth(x):
     return _apply_function(x, 'coth')
 
 
+def sigmoid(x):
+    """Computes coth of x element-wise.
+
+    # Arguments
+        x: Functional object.
+
+    # Returns
+        A new functional object.
+    """
+    return _apply_function(x, 'sigmoid')
+
+
 def abs(x):
     """Computes abs of x element-wise.
 
@@ -444,7 +553,7 @@ def abs(x):
 
 
 def sign(x):
-    """Computes abs of x element-wise.
+    """Computes sign of x element-wise.
 
     # Arguments
         x: Functional object.
@@ -455,6 +564,37 @@ def sign(x):
     return _apply_function(x, 'sign')
 
 
+def step(x):
+    """Computes step (Heaviside) of x element-wise.
+       H(x) = 0 if x<=0
+       H(x) = 1 if x>0
+
+    # Arguments
+        x: Functional object.
+
+    # Returns
+        A new functional object.
+    """
+    validate_functional(x)
+
+    lmbd = []
+    for i in range(len(x.outputs)):
+        lmbd.append(
+            Lambda(
+                lambda x: K.cast(K.greater(x, 0.0), x.dtype), 
+                name=graph_unique_name("step")
+            )
+        )
+        
+    Functional = x.get_class()
+    res = Functional(
+        inputs = x.inputs.copy(),
+        outputs = _apply_operation(lmbd, x),
+        layers = lmbd
+    )
+    return res
+
+
 def log(x):
     """Computes log of x element-wise.
 
@@ -467,6 +607,18 @@ def log(x):
     return _apply_function(x, 'log')
 
 
+def log10(x):
+    """Computes log10 of x element-wise.
+
+    # Arguments
+        x: Functional object.
+
+    # Returns
+        A new functional object.
+    """
+    return log(x)/np.log(10.0)
+
+
 def exp(x):
     """Computes exp of x element-wise.
 
@@ -541,14 +693,22 @@ def _apply_function(x, fname, **kwargs):
         A new functional object.
     """
     validate_functional(x)
-    res = x.copy()
 
     fun = get_activation(fname)
-    lmbd = [Lambda(lambda x: fun(x, **kwargs)) for X in x.outputs]
-    for l in lmbd:
-        l.name = "{}/".format(fname) + l.name.split("_")[-1]
-    # res.append_to_layers(lmbd)
-    res.outputs = _apply_operation(lmbd, x)
+    lmbd = []
+    for i in range(len(x.outputs)):
+        lmbd.append(
+            Lambda(
+                lambda x: fun(x, **kwargs),
+                name=graph_unique_name("{}".format(fname))
+            )
+        )
+    Functional = x.get_class()
+    res = Functional(
+        inputs = x.inputs.copy(),
+        outputs = _apply_operation(lmbd, x),
+        layers = lmbd
+    )
     return res
 
 
@@ -562,44 +722,159 @@ def getitem(x, item):
         A new functional object.
     """
     validate_functional(x)
-    res = x.copy()
 
+    in_item = item
+    print(in_item)
+    if not isinstance(in_item, tuple):
+        in_item = (in_item,)
+    print(in_item)
+
+    itms = (slice(None, None, None),)
+    for it in in_item:
+        itms += (slice(it, it+1) if isinstance(it, int) else it, )
+    
+    lmbd = []
     ys = []
-    lmbd = [Lambda(lambda xx: K.slice(xx, [0, item], [-1, 1])) for xx in x.outputs]
-    for l, y in zip(lmbd, x.outputs):
-        # l.name = "slice/" + l.name.split("_")[-1]
+    for y in x.outputs:
+        l = Lambda(
+            lambda xx: xx[itms], 
+            name=graph_unique_name("slice")
+        )
+        lmbd.append(l)
         ys.append(l(y))
 
-    res.outputs = ys
+    Functional = x.get_class()
+    res = Functional(
+        inputs = x.inputs.copy(),
+        outputs = ys,
+        layers = lmbd
+    )
     return res
 
 
-def gradients(ys, xs, order=1):
+def _gradients(ys, xs, order=1):
     """Returns the gradients of y in `ys` w.r.t. x in `xs`.
-    
+
     `ys` and `xs` are each a Tensor or a list of tensors.
 
     # Arguments
         ys: A tensor or list of tesnors to be differentiated.
         xs: A tensor or list of tensors to be used for differentiation.
-        order: Order of differentiation. 
+        order: Order of differentiation.
 
     # Returns
         A list of `D^n y / Dx^n` for each y and x in `ys` and `xs`.
     """
-    ds = ys
-    for i in range(order):
-        ds = unpack_singleton(
-            tf.gradients(
-                ds, xs,
-                unconnected_gradients='zero',
-                # colocate_gradients_with_ops=True, TF: V1.14.0
+    if ys.shape[-1] == 1:
+        ds = ys
+        for i in range(order):
+            ds = unpack_singleton(
+                tf_gradients(
+                    ds, xs,
+                    unconnected_gradients='zero',
+                    # colocate_gradients_with_ops=True, TF: V1.14.0
+                )
             )
-        )
+
+    else:
+        splitted_ys = tf.split(ys, num_or_size_splits=ys.shape[-1], axis=-1)
+        ds = []
+        for j, y in enumerate(splitted_ys):
+            ds.append(y)
+            for i in range(order):
+                ds[-1] = unpack_singleton(
+                    tf_gradients(
+                        ds[-1], xs,
+                        unconnected_gradients='zero',
+                        # colocate_gradients_with_ops=True, TF: V1.14.0
+                    )
+                )
+            new_shape = [x if x is not None else -1 for x in ds[-1].shape + (1,)]
+            ds[-1] = K.reshape(ds[-1], new_shape)
+        # The output is a tensor.
+        ds = K.concatenate(ds, -1)
     return ds
 
 
-def lambda_gradient(ys, xs, order=1, name=''):
+def _diag_gradients(ys, xs, order=1):
+    """Returns the gradients of y in `ys` w.r.t. x in `xs`.
+
+    `ys` and `xs` are each a Tensor or a list of tensors.
+
+    # Arguments
+        ys: A tensor or list of tesnors to be differentiated.
+        xs: A tensor or list of tensors to be used for differentiation.
+        order: Order of differentiation.
+
+    # Returns
+        A list of `D^n y / Dx^n` for each y and x in `ys` and `xs`.
+    """
+    assert ys.shape.as_list() == xs.shape.as_list(), \
+        'Supported when X and Y has the same dimensions - ' + \
+        'Xs:{}, Ys:{}'.format(xs.shape.as_list(), ys.shape.as_list())
+
+    ds = _gradients(ys, xs, order)
+    return tf.linalg.diag_part(ds)
+
+
+def _diag_gradients2(ys, xs, order=1):
+    """Returns the gradients of y in `ys` w.r.t. x in `xs`.
+
+    `ys` and `xs` are each a Tensor or a list of tensors.
+
+    # Arguments
+        ys: A tensor or list of tesnors to be differentiated.
+        xs: A tensor or list of tensors to be used for differentiation.
+        order: Order of differentiation.
+
+    # Returns
+        A list of `D^n y / Dx^n` for each y and x in `ys` and `xs`.
+    """
+    assert ys.shape.as_list() == xs.shape.as_list(), \
+        'Supported when X and Y has the same dimensions - ' + \
+        'Xs:{}, Ys:{}'.format(xs.shape.as_list(), ys.shape.as_list())
+
+    splitted_ys = tf.split(ys, num_or_size_splits=ys.shape[-1], axis=-1)
+    ds = []
+    for j, y in enumerate(splitted_ys):
+        ds.append(y)
+        for i in range(order):
+            ds[-1] = unpack_singleton(
+                tf.gradients(
+                    ds[-1], xs,
+                    unconnected_gradients='zero',
+                    # colocate_gradients_with_ops=True, TF: V1.14.0
+                )
+            )
+            
+        ds[-1] = ds[-1][:, j:j+1]
+    # The output is a tensor.
+    ds = K.concatenate(ds, -1)
+    return ds
+
+
+def _div_gradients(ys, xs, order=1):
+    """Returns the gradients of y in `ys` w.r.t. x in `xs`.
+
+    `ys` and `xs` are each a Tensor or a list of tensors.
+
+    # Arguments
+        ys: A tensor or list of tesnors to be differentiated.
+        xs: A tensor or list of tensors to be used for differentiation.
+        order: Order of differentiation.
+
+    # Returns
+        A list of `D^n y / Dx^n` for each y and x in `ys` and `xs`.
+    """
+    assert ys.shape.as_list() == xs.shape.as_list(), \
+        'Supported when X and Y has the same dimensions - ' + \
+        'Xs:{}, Ys:{}'.format(xs.shape.as_list(), ys.shape.as_list())
+
+    ds = _diag_gradients(ys, xs, order)
+    return tf.math.reduce_sum(ds, [1], keepdims=True)
+
+
+def _lambda_gradient(ys, xs, order=1, gtype='Grad', name=''):
     """Returns the gradients of y in `ys` w.r.t. x in `xs` using Lambda layers.
     
     `ys` and `xs` are each a Tensor or a list of tensors.
@@ -607,6 +882,10 @@ def lambda_gradient(ys, xs, order=1, name=''):
     # Arguments
         ys: A tensor or list of tesnors to be differentiated.
         xs: A tensor or list of tensors to be used for differentiation.
+        gtype: type of differentiation - can be:
+            - 'Grad' for gradient operation, i.e. Gij = dy_j / dx_i
+            - 'dGrad' for the diagonal of gradient tensor, i.e. Gi = dy_i / dx_i
+            - 'Div' for divergence operation, i.e. G = sum(dy_i / dx_i)
         name: A str name for the Lambda layer. 
 
     # Returns
@@ -614,22 +893,34 @@ def lambda_gradient(ys, xs, order=1, name=''):
         layers: A Lambda layer or list of Lambda layers where the gradient operator is applied.
         grads: A gradient tensor or list of gradient tensors. 
     """
-    name_prefix = 'Grad_' if order==1 else 'Grad{:d}_'.format(order)
     
     grads, layers = [], []
     for y in to_list(ys):
-        lay = Lambda(lambda y: gradients(y, xs, order))
-        lay.name = name_prefix + name + '/' + lay.name
+        if gtype == 'Grad':
+            name_prefix = 'Grad_' if order == 1 else 'Grad{:d}_'.format(order) + name + '/'
+            lay = Lambda(lambda y: _gradients(y, xs, order), name=graph_unique_name(name_prefix))
+        elif gtype == 'dGrad':
+            name_prefix = 'DiagGrad_' if order == 1 else 'Grad{:d}_'.format(order) + name + '/'
+            lay = Lambda(lambda y: _diag_gradients(y, xs, order), name=graph_unique_name(name_prefix))
+        elif gtype == 'Div':
+            name_prefix = 'Div_' if order == 1 else 'Grad{:d}_'.format(order) + name + '/'
+            lay = Lambda(lambda y: _div_gradients(y, xs, order), name=graph_unique_name(name_prefix))
+        else:
+            raise ValueError(
+                'Unrecognised gradient type: {} \n'.format(type) +
+                '     Please choose among (Grad, dGrad, Div). '
+            )
         layers += to_list(lay)
         grads += to_list(lay(y))
 
     return (unpack_singleton(layers), unpack_singleton(grads))
 
 
-def diff(f, *args, **kwargs):
-    """Computes diff of functional object f.
+def _gdiff(gtype, f, *args, **kwargs):
+    """Computes gradient of functional object f.
 
     # Arguments
+        gtype: gradient type - choose from (Grad, dGrad, Div)
         f: Functional object.
         ys: layer name for `ys` to differentiate.
         xs: layer name for `xs` to be differentiated w.r.t.
@@ -650,6 +941,7 @@ def diff(f, *args, **kwargs):
         )
 
     try:
+        inputs = f.inputs.copy()
         if len(args) == 0:
             x = unpack_singleton(f.inputs)
             assert is_tensor(x), \
@@ -661,6 +953,7 @@ def diff(f, *args, **kwargs):
                 x_name = args[x_id]
                 x = next(l for l in f.layers if l.name == x_name).output
             elif is_functional(args[x_id]):
+                inputs += to_list(args[x_id].inputs)
                 x_lay = args[x_id].layers[-1]
                 x_name = x_lay.name
                 x = x_lay.output
@@ -691,15 +984,93 @@ def diff(f, *args, **kwargs):
     order = 1
     if 'order' in kwargs.keys():
         order = kwargs['order']
-
-    res = f.copy()
     
-    lay, tens = lambda_gradient(
-        y, x, order, "{}_{}".format(y_name, x_name)
+    lay, tens = _lambda_gradient(
+        y, x, order, gtype, "{}_{}".format(y_name, x_name)
     )
+    
+    Functional = type(f)
+    res = Functional(
+        inputs = unique_tensors(inputs),
+        outputs = to_list(tens),
+        layers = to_list(lay)
+    )
+    
+    return res
 
-    res.append_to_layers(to_list(lay))
-    res.outputs = to_list(tens)
+
+def grad(f, *args, **kwargs):
+    """Computes gradient tensor of functional object f.
+
+    # Arguments
+        f: Functional object.
+        ys: layer name for `ys` to differentiate.
+        xs: layer name for `xs` to be differentiated w.r.t.
+        order: order of differentiation w.r.t. xs - defaulted to 1.
+
+    # Returns
+        A new functional object.
+    """
+    return _gdiff("Grad", f, *args, **kwargs)
+
+
+# overlaod for backward compatibility 
+diff = grad
+
+
+# consistency with older versions.
+def diag_grad(f, *args, **kwargs):
+    """Computes diag of gradient tensor of functional object f.
+
+    # Arguments
+        f: Functional object.
+        ys: layer name for `ys` to differentiate.
+        xs: layer name for `xs` to be differentiated w.r.t.
+        order: order of differentiation w.r.t. xs - defaulted to 1.
+
+    # Returns
+        A new functional object.
+    """
+    return _gdiff("dGrad", f, *args, **kwargs)
+
+
+# consistency with older versions.
+def div_grad(f, *args, **kwargs):
+    """Computes Divergence of functional object f.
+
+    # Arguments
+        f: Functional object.
+        ys: layer name for `ys` to differentiate.
+        xs: layer name for `xs` to be differentiated w.r.t.
+        order: order of differentiation w.r.t. xs - defaulted to 1.
+
+    # Returns
+        A new functional object.
+    """
+    return _gdiff("Div", f, *args, **kwargs)
+
+
+def stop_gradient(f):
+    """Equivalent to stop_grdient in tf.
+    Use-case: to force zero-gradient with respect to a part of the network.
+
+    # Arguments
+        f: Functional object.
+
+    # Returns
+        A new functional object.
+    """
+    if is_functional(f):
+        lmbd = [Lambda(lambda x: tf_stop_gradient(x), name=graph_unique_name("stop_grad")) for X in f.outputs]
+    else:
+        raise ValueError('Expected one functional in the arguments.')
+
+    Functional = f.get_class()
+    res = Functional(
+        inputs = unique_tensors(f.inputs.copy()),
+        outputs = _apply_operation(lmbd, f),
+        layers = lmbd,
+    )
 
     return res
 
@@ -751,3 +1122,13 @@ def radial_basis2(xs, ci, radii):
     d /= radii
     
     return exp(-sum([(x - c)**2 for x, c in zip(xs, ci)])/radii**2)
+
+
+def _warn_for_ndarray(other):
+    if isinstance(other, np.ndarray):
+        Warning(
+            'Expecting `Tensor` objects instead of `ndarray`: ' +
+            'Note data should go to the training process and ' +
+            'this operation may break batch training. '
+        )
+
